using UnityEngine;
using System.Collections;
using System;

[AddComponentMenu("Physics/Mesh Deformer")]
public class crump : MonoBehaviour
{
    public MeshFilter filter;
    public float resistence = 0.5f;
    public bool deformCollider = true;
    public float updateFrequency = 0;
    public float vertexMovementLimit = 0;
    public Texture2D resistenceMap;

    private Mesh mesh;
    private MeshCollider meshCollider;
    private Vector3[] originalValues;
    private float sizeMod;

    private Vector3[] workingValues;

    private float[] workingMap;
    private Texture2D curMap;

    private Vector3[][] savedDeforms;

    private bool update = false;
    private float lastUpdate;


    public void Awake()
    {
        meshCollider = GetComponent<MeshCollider>();

        if (filter == null)
        {
            filter = GetComponent<MeshFilter>();
        }

        LoadMesh();
    }

    public void OnCollissionEnter(Collision col)
    {
        DoDeform(col);
    }

    public void OnCollisionStay(Collision col)
    {
        DoDeform(col);
    }

    private void DoDeform(Collision col)
    {
        Debug.Log("Deforming");
        float force = Mathf.Min(1, col.impactForceSum.sqrMagnitude / 1000);
        
        if(force < 0.001)
        {
            return;
        }

        float distF = force * (sizeMod * (0.1f / Mathf.Max(0.1f, resistence)));
        
        foreach(ContactPoint contact in col.contacts)
        {
            for(int i = 0; i < workingValues.Length; i++)
            {
                Vector3 workingVertex = workingValues[i];

                Vector3 localPoint = filter.transform.InverseTransformPoint(contact.point);
                float distM = (workingVertex - localPoint).sqrMagnitude;
                if(distM <= distF)
                {
                    Vector3 norm = filter.transform.InverseTransformDirection(contact.normal * (1 - (distM / distF)) * distF);

                    if(workingMap != null)
                    {
                        norm *= 1 - workingMap[i];
                    }

                    workingVertex += norm;

                    if(vertexMovementLimit > 0)
                    {
                        Vector3 nTwo = workingVertex - originalValues[i];
                        if(nTwo.magnitude > vertexMovementLimit)
                        {
                            workingVertex = originalValues[i] + (nTwo * (vertexMovementLimit / nTwo.magnitude));
                        }
                    }
                }
            }
        }
        Debug.Log("About to request");
        RequestMeshUpdate();
        Debug.Log("Should requested");
    }

    public void FullRestore()
    {
        Restore(1);
    }

    public void Restore(float percent)
    {
        if (percent > 1)
        {
            percent /= 100;
        }
        Restore(percent, Vector3.zero, 0);
    }

    public void Restore(float percent, Vector3 centre, float radius)
    {
        Vector3 localCentre = filter.transform.InverseTransformPoint(centre);

        for (int i = 0; i < workingValues.Length; i++)
        {
            Vector3 workingVertex = workingValues[i];
            if (radius > 0)
            {
                if (!((localCentre - workingVertex).magnitude >= radius))
                {
                    Vector3 diff = workingVertex - originalValues[i];
                    workingVertex = originalValues[i] + (diff * (1 - percent));
                }
            }
        }
        RequestMeshUpdate();
    }

    private void LoadMesh()
    {
        mesh = filter.mesh;
        originalValues = mesh.vertices;
        workingValues = mesh.vertices;

        Vector3 size = mesh.bounds.size;
        sizeMod = Mathf.Min(size.x, size.y, size.z);
    }

    private void LoadMap()
    {
        if (resistenceMap != null)
        {
            curMap = resistenceMap;

            workingMap = new float[mesh.uv.Length];

            int i = 0;
            foreach (Vector2 uv in mesh.uv)
            {
                workingMap[i] = curMap.GetPixelBilinear(uv.x, uv.y).a;
                i++;
            }
        }
        else
        {
            curMap = null;
            workingMap = null;
        }
    }

    private void UpdateMesh()
    {
        Debug.Log("Updating");
        mesh.vertices = workingValues;
        mesh.RecalculateNormals();
        mesh.RecalculateBounds();
        if (deformCollider && (meshCollider != null))
        {
            meshCollider.sharedMesh = null;
            meshCollider.sharedMesh = mesh;
        }
        update = false;
        lastUpdate = Time.time;
    }

    void FixedUpdate()
    {
        if (mesh != filter.mesh)
        {
            LoadMesh();
        }
        if (resistenceMap != curMap)
        {
            LoadMap();
        }
        if (update && (Time.time - lastUpdate >= (1 / updateFrequency)))
        {
            UpdateMesh();
        }
    }

    private void RequestMeshUpdate()
    {
        Debug.Log("Requested Update");
        if (updateFrequency == 0)
        {
            UpdateMesh();
        }
        else
        {
            update = true;
        }
    }

    public Vector3[] GetWorkingVertices()
    {
        return workingValues;
    }

    public void SetWorkingVertices(Vector3[] vertices)
    {
        try
        {
            workingValues = vertices;
        }
        catch(Exception ex)
        {
            Debug.Log(ex.Message);
        }
        finally
        {
            RequestMeshUpdate();
            LoadMap();
        }
    }

    public void LoadDeform(Vector3[] vertices)
    {
        SetWorkingVertices(vertices);
    }

    public void LoadDeform(int saveNumber)
    {
        try
        {
            workingValues = new Vector3[savedDeforms[saveNumber].Length];
            workingValues = savedDeforms[saveNumber];
        }
        catch (Exception ex)
        {
            Debug.Log(ex.Message);
        }
        finally
        {
            RequestMeshUpdate();
            LoadMap();
        }
    }

    public int SaveCurrentDeform()
    {
        int ret = savedDeforms.Length;
        Array.Resize<Vector3[]>(ref savedDeforms, ret + 1);
        savedDeforms[ret] = workingValues;
        return ret;
    }
}


